# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from decimal import Decimal

from trytond.model import fields
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Eval


class Template(metaclass=PoolMeta):
    __name__ = 'product.template'

    sale_quantity_minimal = fields.Float(
        "Sale Minimal Quantity",
        digits=(16, Eval('sale_quantity_digits', 0)),
        states={
            'invisible': ~Eval('salable', False),
            },
        domain=['OR',
            ('sale_quantity_minimal', '=', None),
            ('sale_quantity_minimal', '>', 0),
            ])
    sale_quantity_rounding = fields.Float(
        "Sale Rounding Quantity",
        digits=(16, Eval('sale_quantity_digits', 0)),
        states={
            'invisible': ~Eval('salable', False),
            },
        domain=['OR',
            ('sale_quantity_rounding', '=', None),
            ('sale_quantity_rounding', '>', 0),
            ])
    sale_quantity_digits = fields.Function(
        fields.Integer("Sale Quantity Digits"),
        'on_change_with_sale_quantity_digits')

    @fields.depends('default_uom', 'sale_uom')
    def on_change_with_sale_quantity_digits(self, name=None):
        pool = Pool()
        Uom = pool.get('product.uom')
        if self.default_uom:
            rounding = self.default_uom.rounding
            if self.sale_uom:
                rounding = Uom.compute_qty(
                    self.default_uom, rounding, self.sale_uom, round=False)
            return -Decimal(str(rounding)).as_tuple().exponent


class Product(metaclass=PoolMeta):
    __name__ = 'product.product'



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.tests.test_tryton import ModuleTestCase


class SaleProductQuantityTestCase(ModuleTestCase):
    'Test Sale Product Quantity module'
    module = 'sale_product_quantity'
    extras = ['sale_point']


del ModuleTestCase



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.pool import Pool

from . import product, sale

__all__ = ['register']


def register():
    Pool.register(
        product.Template,
        product.Product,
        sale.Line,
        module='sale_product_quantity', type_='model')
    Pool.register(
        sale.POSLine,
        module='sale_product_quantity', type_='model', depends=['sale_point'])



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.i18n import gettext
from trytond.model import ModelStorage, fields
from trytond.modules.sale.exceptions import SaleValidationError
from trytond.pool import Pool, PoolMeta


def product_quantity_mixin(states):
    class ProductQuantityMixin(ModelStorage):

        @fields.depends('product', 'quantity', 'unit')
        def on_change_product(self):
            pool = Pool()
            Uom = pool.get('product.uom')
            super().on_change_product()
            if self.product and self.quantity is None:
                if self.unit:
                    unit = self.unit
                else:
                    unit = self.product.sale_uom
                if self.product.sale_quantity_minimal:
                    self.quantity = Uom.compute_qty(
                        self.product.sale_uom,
                        self.product.sale_quantity_minimal,
                        unit)

        @classmethod
        def validate(cls, lines):
            super().validate(lines)
            for line in lines:
                line.check_quantity()

        def check_quantity(self):
            pool = Pool()
            Lang = pool.get('ir.lang')
            Uom = pool.get('product.uom')
            if (not self.product
                    or self.sale.state not in states
                    or self.quantity < 0):
                return
            lang = Lang.get()
            quantity = Uom.compute_qty(
                self.unit, self.quantity, self.product.sale_uom, round=False)
            if self.product.sale_quantity_minimal:
                if quantity < self.product.sale_quantity_minimal:
                    minimum = Uom.compute_qty(
                        self.product.sale_uom,
                        self.product.sale_quantity_minimal,
                        self.unit)
                    raise SaleValidationError(gettext(
                            'sale_product_quantity'
                            '.msg_sale_line_quantity_minimal',
                            line=self.rec_name,
                            quantity=lang.format_number_symbol(
                                self.quantity, self.unit),
                            minimum=lang.format_number_symbol(
                                minimum, self.unit)))
            if self.product.sale_quantity_rounding:
                remainder = round(
                    quantity % self.product.sale_quantity_rounding,
                    self.product.sale_quantity_digits)
                if (remainder
                        and remainder != self.product.sale_quantity_rounding):
                    rounding = Uom.compute_qty(
                        self.product.sale_uom,
                        self.product.sale_quantity_rounding,
                        self.unit, round=False)
                    raise SaleValidationError(gettext(
                            'sale_product_quantity'
                            '.msg_sale_line_quantity_rounding',
                            line=self.rec_name,
                            quantity=lang.format_number_symbol(
                                self.quantity, self.unit),
                            rounding=lang.format_number_symbol(
                                rounding, self.unit)))

    return ProductQuantityMixin


class Line(product_quantity_mixin(['draft']), metaclass=PoolMeta):
    __name__ = 'sale.line'


class POSLine(product_quantity_mixin(['open']), metaclass=PoolMeta):
    __name__ = 'sale.point.sale.line'
